;; Generated package description from sql-cassandra.el  -*- no-byte-compile: t -*-
(define-package "sql-cassandra" "0.2.2" "Cassandra support for sql.el" '((emacs "29")) :commit "3e997c6ad82f4240bcd62630ad44b5f023bd16c1" :url "https://elpa.gnu.org/packages/sql-cassandra.html" :authors '(("Filipp Gunbin" . "fgunbin@fastmail.fm")) :maintainer '("Filipp Gunbin" . "fgunbin@fastmail.fm") :keywords '("sql" "cassandra" "cql" "cqlsh"))
